"""Represents tmux integration objects."""
import abc
import typing

import iterm2.api_pb2
import iterm2.connection
import iterm2.rpc
import iterm2.session
import iterm2.transaction
import iterm2.tab
import iterm2.window


@abc.abstractmethod
class Delegate:
    """Delegate interface for tmux."""
    async def tmux_delegate_async_get_window_for_tab_id(
            self, tab_id: str) -> typing.Optional[iterm2.window.Window]:
        """Refreshes and gets the window for the specified tab."""

    def tmux_delegate_get_session_by_id(
            self, session_id: str) -> typing.Optional[iterm2.session.Session]:
        """Returns the session with the given ID."""

    def tmux_delegate_get_connection(self) -> iterm2.connection.Connection:
        """Returns the connection."""


DELEGATE: typing.Optional[Delegate] = None
DELEGATE_FACTORY: typing.Optional[
    typing.Callable[
        [iterm2.connection.Connection],
        typing.Awaitable[Delegate]]] = None


class TmuxException(Exception):
    """A problem was encountered in a Tmux request."""


class TmuxConnection:
    """
    A tmux integration connection.

    Do not create this yourself. Use :func:`async_get_tmux_connections`,
    instead.
    """
    def __init__(self, connection_id, owning_session_id, delegate):
        self.__connection_id = connection_id
        self.__owning_session_id = owning_session_id
        self.__delegate = delegate

    @property
    def connection_id(self) -> str:
        """Returns the unique identifier of the connection.

        :returns: A connection ID, which is also a human-readable description
            of the connection.
        """
        return self.__connection_id

    @property
    def owning_session(self) -> typing.Optional['iterm2.session.Session']:
        """Returns the "gateway" session.

        :returns: The :class:`~iterm2.Session` where `tmux -CC` was run, or
            `None` if it cannot be found."""
        return self.__delegate.tmux_delegate_get_session_by_id(
            self.__owning_session_id)

    async def async_send_command(self, command: str) -> str:
        """Sends a command to the tmux server.

        This may not be called from within a :class:`~iterm2.Transaction`.

        :param command: The command to send to tmux (e.g., "list-sessions")

        :returns: The command's output as a string.

        :throws TmuxException: If the command fails for any reason or the RPC
            fails such as from an invalid ID.

        .. seealso:: Example ":ref:`tile_example`"
        """
        response = await iterm2.rpc.async_rpc_send_tmux_command(
            self.__delegate.tmux_delegate_get_connection(),
            self.__connection_id,
            command)
        # pylint: disable=no-member
        if (response.tmux_response.status == iterm2.api_pb2.TmuxResponse.
                Status.Value("OK")):
            if response.tmux_response.send_command.HasField("output"):
                return response.tmux_response.send_command.output
            raise TmuxException("Tmux reported an error")
        raise TmuxException(
            iterm2.api_pb2.TmuxResponse.Status.Name(
                response.tmux_response.status))

    async def async_set_tmux_window_visible(
            self, tmux_window_id: str, visible: bool) -> None:
        """Hides or shows a tmux window.

        Tmux windows are represented as tabs in iTerm2. You can get a
        tmux_window_id from :meth:`~iterm2.Tab.tmux_window_id`. If this tab is
        attached to a tmux session, then it may be hidden.

        This may not be called from within a :class:`~iterm2.Transaction`.

        :param tmux_window_id: The window to show or hide.
        :param visible: `True` to show a window, `False` to hide a window.
        """
        response = await iterm2.rpc.async_rpc_set_tmux_window_visible(
            self.__delegate.tmux_delegate_get_connection(),
            self.__connection_id,
            tmux_window_id,
            visible)
        # pylint: disable=no-member
        if (response.tmux_response.status != iterm2.api_pb2.TmuxResponse.
                Status.Value("OK")):
            raise TmuxException(
                iterm2.api_pb2.TmuxResponse.Status.Name(
                    response.tmux_response.status))

    async def async_create_window(self) -> 'iterm2.window.Window':
        """Creates a new tmux window.

        This may not be called from within a :class:`~iterm2.Transaction`.

        :returns: A new :class:`Window`.

        .. seealso:: Example ":ref:`tmux_example`"
        """
        response = await iterm2.rpc.async_rpc_create_tmux_window(
            self.__delegate.tmux_delegate_get_connection(),
            self.__connection_id)
        # pylint: disable=no-member
        if (response.tmux_response.status != iterm2.api_pb2.TmuxResponse.
                Status.Value("OK")):
            raise TmuxException(
                iterm2.api_pb2.TmuxResponse.Status.Name(
                    response.tmux_response.status))
        tab_id = response.tmux_response.create_window.tab_id
        return await self.__delegate.tmux_delegate_async_get_window_for_tab_id(
            tab_id)


async def async_get_tmux_connections(
        connection: iterm2.connection.Connection) -> typing.List[
            TmuxConnection]:
    """Fetches a list of tmux connections.

    This may not be called from within a :class:`~iterm2.Transaction`.

    :param connection: The connection to iTerm2.
    :returns: The current tmux connections.

    .. seealso:: Example ":ref:`tmux_example`"
    """
    # Work around a bad design that we're stuck with because this is a public
    # API. Before, tmux depended on app. That was cyclic. The cycle was broken
    # by adding a delegate, but you can call this before creating App. The
    # solution is to use a delegate factory so we can make an App just in time
    # without tmux knowing about it.
    global DELEGATE  # pylint: disable=global-statement
    if not DELEGATE:
        assert DELEGATE_FACTORY
        DELEGATE = await DELEGATE_FACTORY(connection)

    response = await iterm2.rpc.async_rpc_list_tmux_connections(connection)
    # pylint: disable=no-member
    if (response.tmux_response.status ==
            iterm2.api_pb2.TmuxResponse.Status.Value("OK")):
        def make_connection(proto):
            return TmuxConnection(proto.connection_id,
                                  proto.owning_session_id,
                                  DELEGATE)
        return list(map(make_connection,
                        response.tmux_response.list_connections.connections))
    raise TmuxException(
        iterm2.api_pb2.TmuxResponse.Status.Name(
            response.tmux_response.status))


async def async_get_tmux_connection_by_connection_id(
        connection: iterm2.connection.Connection,
        connection_id: str) -> typing.Optional[TmuxConnection]:
    """Find a tmux connection by its ID.

    :param connection: The connection to iTerm2.
    :param connection_id: A connection ID for a :class:`TmuxConnection`.

    :returns: Either a :class:`TmuxConnection` or `None`.
    """
    connections = await async_get_tmux_connections(connection)
    for candidate in connections:
        if candidate.connection_id == connection_id:
            return candidate
    return None



"""Status bar customization interfaces."""

import base64
import enum
import json
import typing

import iterm2.api_pb2
import iterm2.capabilities
import iterm2.color
import iterm2.connection
import iterm2.registration
import iterm2.rpc
import iterm2.util


# pylint: disable=too-few-public-methods
# pylint: disable=too-many-arguments
# pylint: disable=too-many-instance-attributes
class Knob:
    """Represents a configuration setting on a status bar."""
    def __init__(self, knob_type, name, placeholder, json_default_value, key):
        self.__name = name
        self.__type = knob_type
        self.__placeholder = placeholder
        self.__json_default_value = json_default_value
        self.__key = key

    def to_proto(self):
        """Returns a protobuf representation."""
        proto = (iterm2.api_pb2.RPCRegistrationRequest.
                 StatusBarComponentAttributes.Knob())
        proto.name = self.__name
        proto.type = self.__type
        proto.placeholder = self.__placeholder
        proto.json_default_value = self.__json_default_value
        proto.key = self.__key
        return proto


class CheckboxKnob:
    """A status bar configuration knob to select a checkbox.

    :param name: Description of the knob.
    :param default_value: Default value (Boolean).
    :param key: A unique string key identifying this knob.
    """
    def __init__(self, name: str, default_value: bool, key: str):
        self.__knob = Knob(
            (iterm2.api_pb2.RPCRegistrationRequest.
             StatusBarComponentAttributes.Knob.Checkbox),
            name,
            "",
            json.dumps(default_value),
            key)

    def to_proto(self):
        """Returns a protobuf representation."""
        return self.__knob.to_proto()


class StringKnob:
    """A status bar configuration knob to select a string.

    :param name: Description of the knob.
    :param placeholder: Placeholder value (shown in gray) for the text field
        when it has no content.
    :param default_value: Default value.
    :param key: A unique string key identifying this knob.
    """
    def __init__(
            self, name: str, placeholder: str, default_value: str, key: str):
        self.__knob = Knob(
            (iterm2.api_pb2.RPCRegistrationRequest.
             StatusBarComponentAttributes.Knob.String),
            name,
            placeholder,
            json.dumps(default_value),
            key)

    def to_proto(self):
        """Returns a protobuf representation."""
        return self.__knob.to_proto()


class PositiveFloatingPointKnob:
    """
    A status bar configuration knob to select a positive floating point
    value.

    :param name: Description of the knob.
    :param default_value: Default value.
    :param key: A unique string key identifying this knob.
    """
    def __init__(self, name: str, default_value: float, key: str):
        self.__knob = Knob(
            (iterm2.api_pb2.RPCRegistrationRequest.
             StatusBarComponentAttributes.Knob.PositiveFloatingPoint),
            name,
            "",
            json.dumps(default_value),
            key)

    def to_proto(self):
        """Returns a protobuf representation."""
        return self.__knob.to_proto()


class ColorKnob:
    """A status bar configuration knob to select color.

    :param name: Description of the knob.
    :param default_value: Default value.
    :param key: A unique string key identifying this knob
    """
    def __init__(self, name: str, default_value: iterm2.color.Color, key: str):
        self.__knob = Knob(
            (iterm2.api_pb2.RPCRegistrationRequest.
             StatusBarComponentAttributes.Knob.Color),
            name,
            "",
            default_value.json, key)

    def to_proto(self):
        """Returns a protobuf representation."""
        return self.__knob.to_proto()


class StatusBarComponent:
    """Describes a script-provided status bar component showing a text value
    provided by a user-provided coroutine.

    :param short_description: Short description shown below the component in
        the picker UI.
    :param detailed_description: Tool tip for the component in the picker UI.
    :param knobs: List of configuration knobs. See the various Knob classes for
        details.
    :param exemplar: Example value to show in the picker UI as the sample
        content of the component.
    :param update_cadence: How frequently in seconds to reload the value, or
        `None` if it does not need to be reloaded on a timer.
    :param identifier: A string uniquely identifying this component. Use a
        backwards domain name. For example, `com.example.calculator` for a
        calculator component provided by example.com.
    :param icons: An array of `StatusBarComponent.Icon` objects. Should contain
        one with scale 1 and one with scale 2, of size 16x17 points. May be
        empty if you don't want an icon.

    .. seealso::
        * Example ":ref:`escindicator_example`"
        * Example ":ref:`jsonpretty_example`"
        * Example ":ref:`mousemode_example`"
        * Example ":ref:`statusbar_example`"
    """
    class Format(enum.Enum):
        """Describes how a status bar component's output is formatted."""
        PLAIN_TEXT = iterm2.api_pb2.RPCRegistrationRequest.StatusBarComponentAttributes.Format.PLAIN_TEXT
        HTML = iterm2.api_pb2.RPCRegistrationRequest.StatusBarComponentAttributes.Format.HTML    #: A very limited subset of HTML

    class Icon:
        """Contains a status bar icon.

        The scale gives the ratio between pixels and points. For example, a
        32x34 image with scale 2 has a size of 16x17 points.

        Status bar icons should be 16x17 points. Use a two point margin around
        the edges. The text baseline is 3.5 points up from the bottom of the
        image.

        :param scale: 2 for a retina (high-DPI) image, or 1 for a regular
            (low-DPI) image.
        :param base64_data: Base64-encoded data with the icon's image in PNG
            format.
        """
        def __init__(self, scale: float, base64_data: str):
            self.__scale = scale
            self.__data = base64.b64decode(base64_data)
            print(self.__data)

        def to_status_bar_icon(self):
            """Returns a protobuf representation."""
            proto = (iterm2.api_pb2.RPCRegistrationRequest.
                     StatusBarComponentAttributes.Icon())
            proto.data = self.__data
            proto.scale = self.__scale
            return proto

    # pylint: disable=dangerous-default-value
    def __init__(
            self,
            short_description: str,
            detailed_description: str,
            knobs: typing.List[Knob],
            exemplar: str,
            update_cadence: typing.Union[float, None],
            identifier: str,
            icons: typing.List[Icon] = [],
            format: Format = Format.PLAIN_TEXT):
        """Initializes a status bar component."""
        self.__short_description = short_description
        self.__detailed_description = detailed_description
        self.__knobs = knobs
        self.__exemplar = exemplar
        self.__update_cadence = update_cadence
        self.__identifier = identifier
        self.__icons = icons
        self.__connection: typing.Optional[iterm2.connection.Connection] = None
        self.__format = format

    def set_fields_in_proto(self, proto):
        """Populates a protobuf from this object's contents."""
        proto.short_description = self.__short_description
        proto.detailed_description = self.__detailed_description
        knob_protos = list(map(lambda k: k.to_proto(), self.__knobs))
        proto.knobs.extend(knob_protos)
        proto.exemplar = self.__exemplar
        proto.unique_identifier = self.__identifier
        icons = list(map(lambda x: x.to_status_bar_icon(), self.__icons))
        proto.icons.extend(icons)
        proto.format = self.__format.value
        if self.__update_cadence is not None:
            proto.update_cadence = self.__update_cadence

    async def async_open_popover(
            self, session_id: str, html: str, size: iterm2.util.Size):
        """Open a popover with a webview.

        :param session_id: The session identifier.
        :param html: A string containing HTML to show.
        :param size: The desired size of the popover, a
            :class:`~iterm2.util.Size`.

        .. seealso:: Example ":ref:`jsonpretty_example`"
        """
        await iterm2.rpc.async_open_status_bar_component_popover(
            self.__connection,
            self.__identifier,
            session_id,
            html,
            size)

    async def async_set_unread_count(
            self, session_id: typing.Optional[str], count: int):
        """
        Sets the unread count that is displayed in the status bar component. If
        0, it is removed.

        Requires iTerm2 version 3.3.2.

        :param session_id: The session identifier, or none to update all
            instances.
        :param count: The number to show, or 0 to remove it.
        :raises: AppVersionTooOld if not supported by this version of iTerm2.
        """
        if not iterm2.capabilities.supports_status_bar_unread_count(
                self.__connection):
            raise iterm2.capabilities.AppVersionTooOld(
                ("Unread count in status bar components is not " +
                 "supported in this version of iTerm2. Please upgrade " +
                 "to use this script."))

        invocation = iterm2.util.invocation_string(
            "iterm2.set_status_bar_component_unread_count",
            {"identifier": self.__identifier,
             "count": count})
        if session_id:
            await iterm2.rpc.async_invoke_method(
                self.__connection, session_id, invocation, -1)
        else:
            assert self.__connection
            await iterm2.async_invoke_function(self.__connection, invocation)

    async def async_register(
            self,
            connection: iterm2.connection.Connection,
            coro,
            timeout: typing.Union[None, float] = None,
            onclick: typing.Optional[
                typing.Callable[
                    [str, typing.Any],
                    typing.Coroutine[typing.Any, typing.Any, None]]] = None):
        """Registers the statusbar component.

        :param connection: A :class:`~iterm2.Connection`.
        :param coro: An async function. Its arguments are reflected upon to
            determine the RPC's signature. Only the names of the arguments are
            used. All arguments should be keyword arguments as any may be
            omitted at call time. It should take a special argument named
            "knobs" that is a dictionary with configuration settings. It may
            return a string or a list of strings. If it returns a list of
            strings then the longest one that fits will be used.
        :param timeout: How long iTerm2 should wait before giving up on this
            function's ever returning. `None` means to use the default timeout.
        :param onclick: A coroutine to run when the user clicks on the status
            bar component. It should take one argument, which is the session_id
            of the session owning the status bar component that was clicked on.

        Example:

          .. code-block:: python

              component = iterm2.StatusBarComponent(
                  short_description = "Session ID",
                  detailed_description = "Show the session's identifier",
                  knobs = [],
                  exemplar = "[session ID]",
                  update_cadence = None,
                  identifier = "com.iterm2.example.statusbar-rpc")

              @iterm2.StatusBarRPC
              async def session_id_status_bar_coro(
                      knobs,
                      session_id = iterm2.Reference("id")):
                  # This status bar component shows the current session ID,
                  # which is useful for debugging scripts.
                  return session_id

              @iterm2.RPC
              async def my_status_bar_click_handler(session_id):
                  # When you click the status bar it opens a popover with the
                  # message "Hello World"
                  await component.async_open_popover(
                          session_id,
                          "Hello world",
                          iterm2.Size(200, 200))

              await component.async_register(
                      connection,
                      session_id_status_bar_coro,
                      onclick = my_status_bar_click_handler)
        """
        self.__connection = connection
        await coro.async_register(connection, self, timeout)
        if onclick:
            magic_name = "__" + self.__identifier.replace(
                ".", "_").replace("-", "_") + "__on_click"

            async def handle_rpc(session_id):
                await onclick(session_id)

            handle_rpc.__name__ = magic_name
            # This is an abuse of the RPC decorator, but it's a simple way to
            # register a function with a modified name.
            # pylint: disable=no-member
            await (iterm2.registration.RPC(handle_rpc).
                   async_register(connection, timeout=timeout))



"""Provides classes that represent iTerm2 windows."""
import abc
import json
import typing

import iterm2.api_pb2
import iterm2.arrangement
import iterm2.connection
import iterm2.profile
import iterm2.rpc
import iterm2.session
import iterm2.tab
import iterm2.transaction
import iterm2.util


class CreateTabException(Exception):
    """Something went wrong creating a tab."""


class CreateWindowException(Exception):
    """A problem was encountered while creating a window."""


class SetPropertyException(Exception):
    """Something went wrong setting a property."""


class GetPropertyException(Exception):
    """Something went wrong fetching a property."""


DELEGATE_FACTORY: typing.Optional[
    typing.Callable[
        [iterm2.connection.Connection],
        typing.Awaitable['Window.Delegate']]] = None


# pylint: disable=too-many-public-methods
class Window:
    """Represents a terminal window.

    Do not create an instance of `Window` by calling the initializer yourself.
    To get a reference to an existing window, use :class:`~iterm2.app.App` and
    query its `windows` property. To create a new window, use
    :meth:`async_create`.
    """

    class Delegate:
        """Delegate for Window"""
        @abc.abstractmethod
        async def window_delegate_get_window_with_session_id(
                self,
                session_id: str) -> typing.Optional['Window']:
            """Gets the Window that contains a Session by ID."""

        @abc.abstractmethod
        async def window_delegate_get_tab_by_id(
                self,
                tab_id: str) -> typing.Optional[iterm2.tab.Tab]:
            """Gets a Tab by ID."""

        @abc.abstractmethod
        async def window_delegate_get_tab_with_session_id(
                self,
                session_id: str) -> typing.Optional[iterm2.tab.Tab]:
            """Returns the Tab containing a Session by ID."""

    delegate: typing.Optional[Delegate] = None

    @staticmethod
    async def async_create(
            connection: iterm2.connection.Connection,
            profile: str = None,
            command: str = None,
            profile_customizations:
            iterm2.profile.LocalWriteOnlyProfile = None) -> typing.Optional[
                'Window']:
        """Creates a new window.

        :param connection: A :class:`~iterm2.connection.Connection`.
        :param profile: The name of the profile to use for the new window.
        :param command: A command to run in lieu of the shell in the new
            session. Mutually exclusive with profile_customizations.
        :param profile_customizations: LocalWriteOnlyProfile giving changes to
            make in profile. Mutually exclusive with command.

        :returns: A new :class:`Window` or `None` if the session ended right
            away.

        :throws: `CreateWindowException` if something went wrong.

        .. seealso:: Example ":ref:`create_window_example`"
        """
        if command is not None:
            lwop = iterm2.profile.LocalWriteOnlyProfile()
            lwop.set_use_custom_command(
                iterm2.profile.Profile.USE_CUSTOM_COMMAND_ENABLED)
            lwop.set_command(command)
            custom_dict = lwop.values
        elif profile_customizations is not None:
            custom_dict = profile_customizations.values
        else:
            custom_dict = None

        result = await iterm2.rpc.async_create_tab(
            connection,
            profile=profile,
            window=None,
            profile_customizations=custom_dict)
        ctr = result.create_tab_response
        # pylint: disable=no-member
        if ctr.status == iterm2.api_pb2.CreateTabResponse.Status.Value("OK"):
            if not Window.delegate:
                return await Window._async_load(connection, ctr.window_id)

            assert Window.delegate
            return await (
                Window.delegate.window_delegate_get_window_with_session_id(
                    ctr.session_id))

        # pylint: disable=no-member
        raise CreateWindowException(
            iterm2.api_pb2.CreateTabResponse.Status.Name(
                result.create_tab_response.status))

    @staticmethod
    async def _async_load(connection, window_id):
        response = await iterm2.rpc.async_list_sessions(connection)
        list_sessions_response = response.list_sessions_response
        for window in list_sessions_response.windows:
            if window.window_id == window_id:
                return Window.create_from_proto(connection, window)
        return None

    @staticmethod
    def create_from_proto(connection, window):
        """Creates a new instance from a protobuf."""
        tabs = []
        for tab in window.tabs:
            root = iterm2.session.Splitter.from_node(tab.root, connection)
            if tab.HasField("tmux_window_id"):
                tmux_window_id = tab.tmux_window_id
            else:
                tmux_window_id = None
            minimized_sessions = [
                iterm2.session.Session(connection, None, summary)
                for summary in (tab.minimized_sessions or [])]
            tabs.append(
                iterm2.tab.Tab(
                    connection,
                    tab.tab_id,
                    root,
                    tmux_window_id,
                    tab.tmux_connection_id,
                    minimized_sessions))

        if not tabs:
            return None

        return iterm2.window.Window(
            connection,
            window.window_id,
            tabs,
            window.frame,
            window.number)

    # pylint: disable=too-many-arguments
    def __init__(self, connection, window_id, tabs, frame, number):
        self.connection = connection
        self.__window_id = window_id
        self.__tabs = tabs
        self.frame = frame
        self.__number = number
        # None means unknown. Can get set later.
        self.selected_tab_id = None

    def __repr__(self):
        return "<Window id=%s tabs=%s frame=%s>" % (
            self.__window_id,
            self.__tabs,
            iterm2.util.frame_str(self.frame))

    def update_from(self, other):
        """Copies state from other window to this one."""
        self.__tabs = other.tabs
        self.frame = other.frame

    def update_tab(self, tab):
        """Replace references to a tab."""
        i = 0
        for old_tab in self.__tabs:
            if old_tab.tab_id == tab.tab_id:
                self.__tabs[i] = tab
                return
            i += 1

    def pretty_str(self, indent: str = "") -> str:
        """
        :returns: A nicely formatted string describing the window, its tabs,
            and their sessions.
        """
        session = indent + "Window id=%s frame=%s\n" % (
            self.window_id, iterm2.util.frame_str(self.frame))
        for tab in self.__tabs:
            session += tab.pretty_str(indent=indent + "  ")
        return session

    @property
    def window_number(self) -> int:
        """
        :returns: The window's number. When less than 10, this is the number
            part of the shortcut to switch to the window.
        """
        return self.__number

    @property
    def window_id(self) -> str:
        """
        :returns: the window's unique identifier.
        """
        return self.__window_id

    @property
    def tabs(self) -> typing.List[iterm2.tab.Tab]:
        """
        :returns: a list of iterm2.tab.Tab objects.
        """
        return self.__tabs

    async def async_set_tabs(self, tabs: typing.List[iterm2.tab.Tab]):
        """Changes the tabs and their order.

        The provided tabs may belong to any window. They will be moved if
        needed. Windows entirely denuded of tabs will be closed.

        All provided tabs will be inserted in the given order starting at the
        first positions. Any tabs already belonging to this window not in the
        list will remain after the provided tabs.

        :param tabs: a list of tabs, forming the new set of tabs in this
            window.
        :throws: RPCException if something goes wrong.

        .. seealso::
            * Example ":ref:`movetab_example`"
            * Example ":ref:`mrutabs_example`"
            * Example ":ref:`sorttabs_example`"
        """
        tab_ids = map(lambda tab: tab.tab_id, tabs)
        await iterm2.rpc.async_reorder_tabs(
            self.connection,
            assignments=[(self.window_id, tab_ids)])

    @property
    def current_tab(self) -> typing.Optional[iterm2.tab.Tab]:
        """
        :returns: The current tab in this window or `None` if it could not be
            determined.
        """
        for tab in self.__tabs:
            if tab.tab_id == self.selected_tab_id:
                return tab
        return None

    async def async_create_tmux_tab(
            self,
            tmux_connection:
            'iterm2.tmux.TmuxConnection') -> typing.Optional[iterm2.tab.Tab]:
        """Creates a new tmux tab in this window.

        This may not be called from within a
        :class:`~iterm2.transaction.Transaction`.

        :param tmux_connection: The tmux connection to own the new tab.

        :returns: A newly created tab, or `None` if it could not be created.

        :throws: `CreateTabException` if something went wrong.

        .. seealso:: Example ":ref:`tmux_example`"
        """
        tmux_window_id = "{}".format(-(self.__number + 1))
        response = await iterm2.rpc.async_rpc_create_tmux_window(
            self.connection,
            tmux_connection.connection_id,
            tmux_window_id)
        # pylint: disable=no-member
        if (response.tmux_response.status !=
                iterm2.api_pb2.TmuxResponse.Status.Value("OK")):
            raise CreateTabException(
                iterm2.api_pb2.TmuxResponse.Status.Name(
                    response.tmux_response.status))
        tab_id = response.tmux_response.create_window.tab_id
        if not Window.delegate:
            assert DELEGATE_FACTORY
            Window.delegate = await DELEGATE_FACTORY(self.connection)
        return await Window.delegate.window_delegate_get_tab_by_id(tab_id)

    async def async_create_tab(
            self,
            profile: typing.Optional[str] = None,
            command: typing.Optional[str] = None,
            index: typing.Optional[int] = None,
            profile_customizations: typing.Optional[
                iterm2.profile.LocalWriteOnlyProfile] = None) -> typing.Optional[
                    iterm2.tab.Tab]:
        """
        Creates a new tab in this window.

        :param profile: The profile name to use or None for the default
            profile.
        :param command: The command to run in the new session, or None for the
            default for the profile. Mutually exclusive with
            profile_customizations.
        :param index: The index in the window where the new tab should go
            (0=first position, etc.)
        :param profile_customizations: LocalWriteOnlyProfile giving changes to
            make in profile. Mutually exclusive with command.

        :returns: :class:`Tab` or `None` if the session closed right away.

        :throws: CreateTabException if something goes wrong.
        """
        if command is not None:
            lwop = iterm2.profile.LocalWriteOnlyProfile()
            lwop.set_use_custom_command(
                iterm2.profile.Profile.USE_CUSTOM_COMMAND_ENABLED)
            lwop.set_command(command)
            custom_dict = lwop.values
        elif profile_customizations is not None:
            custom_dict = profile_customizations.values
        else:
            custom_dict = None
        result = await iterm2.rpc.async_create_tab(
            self.connection,
            profile=profile,
            window=self.__window_id,
            index=index,
            profile_customizations=custom_dict)
        # pylint: disable=no-member
        if (result.create_tab_response.status ==
                iterm2.api_pb2.CreateTabResponse.Status.Value("OK")):
            session_id = result.create_tab_response.session_id
            assert Window.delegate
            return await Window.delegate.window_delegate_get_tab_with_session_id(
                session_id)
        # pylint: disable=no-member
        raise CreateTabException(
            iterm2.api_pb2.CreateTabResponse.Status.Name(
                result.create_tab_response.status))

    async def async_get_frame(self) -> iterm2.util.Frame:
        """
        Gets the window's frame.

        The origin (0,0) is the *bottom* right of the main screen.

        :returns: This window's frame. This includes window decoration such as
            the title bar.

        :throws: :class:`GetPropertyException` if something goes wrong.
        """

        response = await iterm2.rpc.async_get_property(
            self.connection, "frame", self.__window_id)
        status = response.get_property_response.status
        # pylint: disable=no-member
        if status == iterm2.api_pb2.GetPropertyResponse.Status.Value("OK"):
            frame_dict = json.loads(response.get_property_response.json_value)
            frame = iterm2.util.Frame()
            frame.load_from_dict(frame_dict)
            return frame
        raise GetPropertyException(response.get_property_response.status)

    async def async_set_frame(self, frame: iterm2.util.Frame) -> None:
        """
        Sets the window's frame.

        :param frame: The desired frame.

        :throws: :class:`SetPropertyException` if something goes wrong.
        """
        json_value = json.dumps(frame.dict)
        response = await iterm2.rpc.async_set_property(
            self.connection,
            "frame",
            json_value,
            window_id=self.__window_id)
        status = response.set_property_response.status
        # pylint: disable=no-member
        if status != iterm2.api_pb2.SetPropertyResponse.Status.Value("OK"):
            raise SetPropertyException(response.get_property_response.status)

    async def async_get_fullscreen(self) -> bool:
        """
        Checks if the window is full-screen.

        :returns: Whether the window is full screen.

        :throws: :class:`GetPropertyException` if something goes wrong.
        """
        response = await iterm2.rpc.async_get_property(
            self.connection, "fullscreen", self.__window_id)
        status = response.get_property_response.status
        # pylint: disable=no-member
        if status == iterm2.api_pb2.GetPropertyResponse.Status.Value("OK"):
            return json.loads(response.get_property_response.json_value)
        raise GetPropertyException(response.get_property_response.status)

    async def async_set_fullscreen(self, fullscreen: bool):
        """
        Changes the window's full-screen status.

        :param fullscreen: Whether you wish the window to be full screen.

        :throws: :class:`SetPropertyException` if something goes wrong (such as
            that the fullscreen status could not be changed).
        """
        json_value = json.dumps(fullscreen)
        response = await iterm2.rpc.async_set_property(
            self.connection,
            "fullscreen",
            json_value,
            window_id=self.__window_id)
        status = response.set_property_response.status
        # pylint: disable=no-member
        if status != iterm2.api_pb2.SetPropertyResponse.Status.Value("OK"):
            raise SetPropertyException(response.set_property_response.status)

    async def async_activate(self) -> None:
        """
        Gives the window keyboard focus and orders it to the front.
        """
        await iterm2.rpc.async_activate(
            self.connection,
            False,
            False,
            True,
            window_id=self.__window_id)

    async def async_close(self, force: bool = False):
        """
        Closes the window.

        :param force: If `True`, the user will not be prompted for a
            confirmation.

        :throws: :class:`RPCException` if something goes wrong.
        """
        result = await iterm2.rpc.async_close(
            self.connection, windows=[self.__window_id], force=force)
        status = result.close_response.statuses[0]
        # pylint: disable=no-member
        if status != iterm2.api_pb2.CloseResponse.Status.Value("OK"):
            raise iterm2.rpc.RPCException(
                iterm2.api_pb2.CloseResponse.Status.Name(status))

    async def async_save_window_as_arrangement(self, name: str) -> None:
        """Save the current window as a new arrangement.

        :param name: The name to save as. Will overwrite if one already exists
            with this name.
        """
        result = await iterm2.rpc.async_save_arrangement(
            self.connection, name, self.__window_id)
        # pylint: disable=no-member
        if (result.create_tab_response.status !=
                iterm2.api_pb2.CreateTabResponse.Status.Value("OK")):
            raise iterm2.arrangement.SavedArrangementException(
                iterm2.api_pb2.SavedArrangementResponse.Status.Name(
                    result.saved_arrangement_response.status))

    async def async_restore_window_arrangement(self, name: str) -> None:
        """Restore a window arrangement as tabs in this window.

        :param name: The name to restore.

        :throws: :class:`~iterm2.arrangement.SavedArrangementException` if the
            named arrangement does not exist."""
        result = await iterm2.rpc.async_restore_arrangement(
            self.connection, name, self.__window_id)
        # pylint: disable=no-member
        if (result.create_tab_response.status !=
                iterm2.api_pb2.CreateTabResponse.Status.Value("OK")):
            raise iterm2.arrangement.SavedArrangementException(
                iterm2.api_pb2.SavedArrangementResponse.Status.Name(
                    result.saved_arrangement_response.status))

    async def async_get_variable(self, name: str) -> typing.Any:
        """
        Fetches a window variable.

        See the Scripting Fundamentals documentation for more information on
        variables.

        :param name: The variable's name.

        :returns: The variable's value or empty string if it is undefined.

        :throws: :class:`~iterm2.rpc.RPCException` if something goes wrong.
        """
        # pylint: disable=no-member
        result = await iterm2.rpc.async_variable(
            self.connection, window_id=self.__window_id, gets=[name])
        status = result.variable_response.status
        if status != iterm2.api_pb2.VariableResponse.Status.Value("OK"):
            raise iterm2.rpc.RPCException(
                iterm2.api_pb2.VariableResponse.Status.Name(status))
        return json.loads(result.variable_response.values[0])

    async def async_set_variable(self, name: str, value: typing.Any) -> None:
        """
        Sets a user-defined variable in the window.

        See the Scripting Fundamentals documentation for more information on
        user-defined variables.

        :param name: The variable's name. Must begin with `user.`.
        :param value: The new value to assign.

        :throws: :class:`RPCException` if something goes wrong.
        """
        # pylint: disable=no-member
        result = await iterm2.rpc.async_variable(
            self.connection,
            sets=[(name, json.dumps(value))],
            window_id=self.window_id)
        status = result.variable_response.status
        if status != iterm2.api_pb2.VariableResponse.Status.Value("OK"):
            raise iterm2.rpc.RPCException(
                iterm2.api_pb2.VariableResponse.Status.Name(status))

    async def async_set_title(self, title: str):
        """Changes the window's title.

        This is equivalent to editing the window's title with the menu item
        Edit Window Title. The title is an interpolated string. Note that when
        using tmux integration, tab titles correspond to tmux window titles.
        iTerm2's window titles have no equivalent in tmux.

        :param title: The new title. Set it to an empty string to use the
            default value (the current tab's title).

        :throws: :class:`~iterm2.rpc.RPCException` if something goes wrong.
        """
        invocation = iterm2.util.invocation_string(
            "iterm2.set_title",
            {"title": title})
        await iterm2.rpc.async_invoke_method(
            self.connection, self.window_id, invocation, -1)

    async def async_invoke_function(
            self, invocation: str, timeout: float = -1):
        """
        Invoke an RPC. Could be a registered function by this or another script
        of a built-in function.

        This invokes the RPC in the context of this window. Note that most
        user-defined RPCs expect to be invoked in the context of a session.
        Default variables will be pulled from that scope. If you call a
        function from the wrong context it may fail because its defaults will
        not be set properly.

        :param invocation: A function invocation string.
        :param timeout: Max number of secondsto wait. Negative values mean to
            use the system default timeout.

        :returns: The result of the invocation if successful.

        :throws: :class:`~iterm2.rpc.RPCException` if something goes wrong.
        """
        response = await iterm2.rpc.async_invoke_function(
            self.connection,
            invocation,
            window_id=self.window_id,
            timeout=timeout)
        which = response.invoke_function_response.WhichOneof('disposition')
        # pylint: disable=no-member
        if which == 'error':
            if (response.invoke_function_response.error.status ==
                    iterm2.api_pb2.InvokeFunctionResponse.Status.Value(
                        "TIMEOUT")):
                raise iterm2.rpc.RPCException("Timeout")
            raise iterm2.rpc.RPCException("{}: {}".format(
                iterm2.api_pb2.InvokeFunctionResponse.Status.Name(
                    response.invoke_function_response.error.status),
                response.invoke_function_response.error.error_reason))
        return json.loads(
            response.invoke_function_response.success.json_result)



"""Defines interfaces for accessing menu items."""
import enum
import iterm2.api_pb2
import iterm2.rpc
import typing

class MenuItemException(Exception):
    """A problem was encountered while selecting a menu item."""


class MenuItemState:
    """Describes the current state of a menu item."""
    def __init__(self, checked: bool, enabled: bool):
        self.__checked = checked
        self.__enabled = enabled

    @property
    def checked(self):
        """Is the menu item checked? A `bool` property."""
        return self.__checked

    @property
    def enabled(self):
        """
        Is the menu item enabled (i.e., it can be selected)? A `bool`
        property.
        """
        return self.__enabled

class MenuItemIdentifier:
    def __init__(self, title, identifier):
        self.__title = title
        self.__identifier = identifier

    def __repr__(self):
        return f'[MenuItemIdentifier title={self.title} id={self.identifier}]'

    @property
    def title(self) -> str:
        return self.__title

    @property
    def identifier(self) -> typing.Optional[str]:
        return self.__identifier

    def _encode(self):
        # Encodes to a key binding parameter.
        if self.__identifier is None:
            return self.__title
        return self.__title + "\n" + self.__identifier

class MainMenu:
    """Represents the app's main menu."""

    @staticmethod
    async def async_select_menu_item(connection, identifier: str):
        """Selects a menu item.

        :param identifier: A string. See list of identifiers in :doc:`menu_ids`

        :throws MenuItemException: if something goes wrong.

        .. seealso:: Example ":ref:`zoom_on_screen_example`"
        """
        response = await iterm2.rpc.async_menu_item(
            connection, identifier, False)
        status = response.menu_item_response.status
        # pylint: disable=no-member
        if status != iterm2.api_pb2.MenuItemResponse.Status.Value("OK"):
            raise MenuItemException(
                iterm2.api_pb2.MenuItemResponse.Status.Name(status))

    @staticmethod
    async def async_get_menu_item_state(
            connection, identifier: str) -> MenuItemState:
        """Queries a menu item for its state.

        :param identifier: A string. See list of identifiers in :doc:`menu_ids`

        :throws MenuItemException: if something goes wrong.
        """
        response = await iterm2.rpc.async_menu_item(
            connection, identifier, True)
        status = response.menu_item_response.status
        # pylint: disable=no-member
        if status != iterm2.api_pb2.MenuItemResponse.Status.Value("OK"):
            raise MenuItemException(
                iterm2.api_pb2.MenuItemResponse.Status.Name(status))
        return iterm2.MenuItemState(
            response.menu_item_response.checked,
            response.menu_item_response.enabled)

    class iTerm2(enum.Enum):
        ABOUT_ITERM2 = MenuItemIdentifier("About iTerm2", "About iTerm2")
        SHOW_TIP_OF_THE_DAY = MenuItemIdentifier("Show Tip of the Day", "Show Tip of the Day")
        CHECK_FOR_UPDATES = MenuItemIdentifier("Check For Updates…", "Check For Updates…")
        TOGGLE_DEBUG_LOGGING = MenuItemIdentifier("Toggle Debug Logging", "Toggle Debug Logging")
        COPY_PERFORMANCE_STATS = MenuItemIdentifier("Copy Performance Stats", "Copy Performance Stats")
        CAPTURE_GPU_FRAME = MenuItemIdentifier("Capture GPU Frame", "Capture Metal Frame")
        PREFERENCES = MenuItemIdentifier("Preferences...", "Preferences...")
        HIDE_ITERM2 = MenuItemIdentifier("Hide iTerm2", "Hide iTerm2")
        HIDE_OTHERS = MenuItemIdentifier("Hide Others", "Hide Others")
        SHOW_ALL = MenuItemIdentifier("Show All", "Show All")
        SECURE_KEYBOARD_ENTRY = MenuItemIdentifier("Secure Keyboard Entry", "Secure Keyboard Entry")
        MAKE_ITERM2_DEFAULT_TERM = MenuItemIdentifier("Make iTerm2 Default Term", "Make iTerm2 Default Term")
        MAKE_TERMINAL_DEFAULT_TERM = MenuItemIdentifier("Make Terminal Default Term", "Make Terminal Default Term")
        INSTALL_SHELL_INTEGRATION = MenuItemIdentifier("Install Shell Integration", "Install Shell Integration")
        QUIT_ITERM2 = MenuItemIdentifier("Quit iTerm2", "Quit iTerm2")

    class Shell(enum.Enum):
        NEW_WINDOW = MenuItemIdentifier("New Window", "New Window")
        NEW_WINDOW_WITH_CURRENT_PROFILE = MenuItemIdentifier("New Window with Current Profile", "New Window with Current Profile")
        NEW_TAB = MenuItemIdentifier("New Tab", "New Tab")
        NEW_TAB_WITH_CURRENT_PROFILE = MenuItemIdentifier("New Tab with Current Profile", "New Tab with Current Profile")
        DUPLICATE_TAB = MenuItemIdentifier("Duplicate Tab", "Duplicate Tab")
        SPLIT_HORIZONTALLY_WITH_CURRENT_PROFILE = MenuItemIdentifier("Split Horizontally with Current Profile", "Split Horizontally with Current Profile")
        SPLIT_VERTICALLY_WITH_CURRENT_PROFILE = MenuItemIdentifier("Split Vertically with Current Profile", "Split Vertically with Current Profile")
        SPLIT_HORIZONTALLY = MenuItemIdentifier("Split Horizontally…", "Split Horizontally…")
        SPLIT_VERTICALLY = MenuItemIdentifier("Split Vertically…", "Split Vertically…")
        SAVE_CONTENTS = MenuItemIdentifier("Save Contents…", "Log.SaveContents")
        SAVE_SELECTED_TEXT = MenuItemIdentifier("Save Selected Text…", "Save Selected Text…")
        CLOSE = MenuItemIdentifier("Close", "Close")
        CLOSE_TERMINAL_WINDOW = MenuItemIdentifier("Close Terminal Window", "Close Terminal Window")
        CLOSE_ALL_PANES_IN_TAB = MenuItemIdentifier("Close All Panes in Tab", "Close All Panes in Tab")
        UNDO_CLOSE = MenuItemIdentifier("Undo Close", "Undo Close")

        class BroadcastInput(enum.Enum):
            SEND_INPUT_TO_CURRENT_SESSION_ONLY = MenuItemIdentifier("Send Input to Current Session Only", "Broadcast Input.Send Input to Current Session Only")
            BROADCAST_INPUT_TO_ALL_PANES_IN_ALL_TABS = MenuItemIdentifier("Broadcast Input to All Panes in All Tabs", "Broadcast Input.Broadcast Input to All Panes in All Tabs")
            BROADCAST_INPUT_TO_ALL_PANES_IN_CURRENT_TAB = MenuItemIdentifier("Broadcast Input to All Panes in Current Tab", "Broadcast Input.Broadcast Input to All Panes in Current Tab")
            TOGGLE_BROADCAST_INPUT_TO_CURRENT_SESSION = MenuItemIdentifier("Toggle Broadcast Input to Current Session", "Broadcast Input.Toggle Broadcast Input to Current Session")
            SHOW_BACKGROUND_PATTERN_INDICATOR = MenuItemIdentifier("Show Background Pattern Indicator", "Broadcast Input.Show Background Pattern Indicator")


        class tmux(enum.Enum):
            DETACH = MenuItemIdentifier("Detach", "tmux.Detach")
            FORCE_DETACH = MenuItemIdentifier("Force Detach", "tmux.Force Detach")
            NEW_TMUX_WINDOW = MenuItemIdentifier("New Tmux Window", "tmux.New Tmux Window")
            NEW_TMUX_TAB = MenuItemIdentifier("New Tmux Tab", "tmux.New Tmux Tab")
            PAUSE_PANE = MenuItemIdentifier("Pause Pane", "trmux.Pause Pane")
            DASHBOARD = MenuItemIdentifier("Dashboard", "tmux.Dashboard")

        PAGE_SETUP = MenuItemIdentifier("Page Setup...", "Page Setup...")

        class Print(enum.Enum):
            SCREEN = MenuItemIdentifier("Screen", "Print.Screen")
            SELECTION = MenuItemIdentifier("Selection", "Print.Selection")
            BUFFER = MenuItemIdentifier("Buffer", "Print.Buffer")


    class Edit(enum.Enum):
        UNDO = MenuItemIdentifier("Undo", "Undo")
        REDO = MenuItemIdentifier("Redo", "Redo")
        CUT = MenuItemIdentifier("Cut", "Cut")
        COPY = MenuItemIdentifier("Copy", "Copy")
        COPY_WITH_STYLES = MenuItemIdentifier("Copy with Styles", "Copy with Styles")
        COPY_WITH_CONTROL_SEQUENCES = MenuItemIdentifier("Copy with Control Sequences", "Copy with Control Sequences")
        COPY_MODE = MenuItemIdentifier("Copy Mode", "Copy Mode")
        PASTE = MenuItemIdentifier("Paste", "Paste")

        class PasteSpecial(enum.Enum):
            ADVANCED_PASTE = MenuItemIdentifier("Advanced Paste…", "Paste Special.Advanced Paste…")
            PASTE_SELECTION = MenuItemIdentifier("Paste Selection", "Paste Special.Paste Selection")
            PASTE_FILE_BASE64ENCODED = MenuItemIdentifier("Paste File Base64-Encoded", "Paste Special.Paste File Base64-Encoded")
            PASTE_SLOWLY = MenuItemIdentifier("Paste Slowly", "Paste Special.Paste Slowly")
            PASTE_FASTER = MenuItemIdentifier("Paste Faster", "Paste Special.Paste Faster")
            PASTE_SLOWLY_FASTER = MenuItemIdentifier("Paste Slowly Faster", "Paste Special.Paste Slowly Faster")
            PASTE_SLOWER = MenuItemIdentifier("Paste Slower", "Paste Special.Paste Slower")
            PASTE_SLOWLY_SLOWER = MenuItemIdentifier("Paste Slowly Slower", "Paste Special.Paste Slowly Slower")
            WARN_BEFORE_MULTILINE_PASTE = MenuItemIdentifier("Warn Before Multi-Line Paste", "Paste Special.Warn Before Multi-Line Paste")
            PROMPT_TO_CONVERT_TABS_TO_SPACES_WHEN_PASTING = MenuItemIdentifier("Prompt to Convert Tabs to Spaces when Pasting", "Paste Special.Prompt to Convert Tabs to Spaces when Pasting")
            LIMIT_MULTILINE_PASTE_WARNING_TO_SHELL_PROMPT = MenuItemIdentifier("Limit Multi-Line Paste Warning to Shell Prompt", "Paste Special.Limit Multi-Line Paste Warning to Shell Prompt")
            WARN_BEFORE_PASTING_ONE_LINE_ENDING_IN_A_NEWLINE_AT_SHELL_PROMPT = MenuItemIdentifier("Warn Before Pasting One Line Ending in a Newline at Shell Prompt", "Paste Special.Warn Before Pasting One Line Ending in a Newline at Shell Prompt")

        OPEN_SELECTION = MenuItemIdentifier("Open Selection", "Open Selection")
        JUMP_TO_SELECTION = MenuItemIdentifier("Jump to Selection", "Find.Jump to Selection")
        SELECT_ALL = MenuItemIdentifier("Select All", "Select All")
        SELECTION_RESPECTS_SOFT_BOUNDARIES = MenuItemIdentifier("Selection Respects Soft Boundaries", "Selection Respects Soft Boundaries")
        SELECT_OUTPUT_OF_LAST_COMMAND = MenuItemIdentifier("Select Output of Last Command", "Select Output of Last Command")
        SELECT_CURRENT_COMMAND = MenuItemIdentifier("Select Current Command", "Select Current Command")

        class Find(enum.Enum):
            FIND = MenuItemIdentifier("Find...", "Find.Find...")
            FIND_NEXT = MenuItemIdentifier("Find Next", "Find.Find Next")
            FIND_PREVIOUS = MenuItemIdentifier("Find Previous", "Find.Find Previous")
            USE_SELECTION_FOR_FIND = MenuItemIdentifier("Use Selection for Find", "Find.Use Selection for Find")
            FIND_GLOBALLY = MenuItemIdentifier("Find Globally...", "Find.Find Globally...")
            FIND_URLS = MenuItemIdentifier("Find URLs", "Find.Find URLs")
            FILTER = MenuItemIdentifier("Filter", "Find.Filter")


        class MarksandAnnotations(enum.Enum):
            SET_MARK = MenuItemIdentifier("Set Mark", "Marks and Annotations.Set Mark")
            JUMP_TO_MARK = MenuItemIdentifier("Jump to Mark", "Marks and Annotations.Jump to Mark")
            NEXT_MARK = MenuItemIdentifier("Next Mark", "Marks and Annotations.Next Mark")
            PREVIOUS_MARK = MenuItemIdentifier("Previous Mark", "Marks and Annotations.Previous Mark")
            ADD_ANNOTATION_AT_CURSOR = MenuItemIdentifier("Add Annotation at Cursor", "Marks and Annotations.Add Annotation at Cursor")
            NEXT_ANNOTATION = MenuItemIdentifier("Next Annotation", "Marks and Annotations.Next  Annotation")
            PREVIOUS_ANNOTATION = MenuItemIdentifier("Previous Annotation", "Marks and Annotations.Previous  Annotation")

            class Alerts(enum.Enum):
                ALERT_ON_NEXT_MARK = MenuItemIdentifier("Alert on Next Mark", "Marks and Annotations.Alerts.Alert on Next Mark")
                SHOW_MODAL_ALERT_BOX = MenuItemIdentifier("Show Modal Alert Box", "Marks and Annotations.Alerts.Show Modal Alert Box")
                POST_NOTIFICATION = MenuItemIdentifier("Post Notification", "Marks and Annotations.Alerts.Post Notification")


        CLEAR_BUFFER = MenuItemIdentifier("Clear Buffer", "Clear Buffer")
        CLEAR_SCROLLBACK_BUFFER = MenuItemIdentifier("Clear Scrollback Buffer", "Clear Scrollback Buffer")
        CLEAR_TO_START_OF_SELECTION = MenuItemIdentifier("Clear to Start of Selection", "Clear to Start of Selection")
        CLEAR_TO_LAST_MARK = MenuItemIdentifier("Clear to Last Mark", "Clear to Last Mark")

    class View(enum.Enum):
        SHOW_TABS_IN_FULLSCREEN = MenuItemIdentifier("Show Tabs in Fullscreen", "Show Tabs in Fullscreen")
        TOGGLE_FULL_SCREEN = MenuItemIdentifier("Toggle Full Screen", "Toggle Full Screen")
        USE_TRANSPARENCY = MenuItemIdentifier("Use Transparency", "Use Transparency")
        ZOOM_IN_ON_SELECTION = MenuItemIdentifier("Zoom In on Selection", "Zoom In on Selection")
        ZOOM_OUT = MenuItemIdentifier("Zoom Out", "Zoom Out")
        FIND_CURSOR = MenuItemIdentifier("Find Cursor", "Find Cursor")
        SHOW_CURSOR_GUIDE = MenuItemIdentifier("Show Cursor Guide", "Show Cursor Guide")
        SHOW_TIMESTAMPS = MenuItemIdentifier("Show Timestamps", "Show Timestamps")
        SHOW_ANNOTATIONS = MenuItemIdentifier("Show Annotations", "Show Annotations")
        AUTO_COMMAND_COMPLETION = MenuItemIdentifier("Auto Command Completion", "Auto Command Completion")
        COMPOSER = MenuItemIdentifier("Composer", "Composer")
        OPEN_QUICKLY = MenuItemIdentifier("Open Quickly", "Open Quickly")
        MAXIMIZE_ACTIVE_PANE = MenuItemIdentifier("Maximize Active Pane", "Maximize Active Pane")
        MAKE_TEXT_BIGGER = MenuItemIdentifier("Make Text Bigger", "Make Text Bigger")
        MAKE_TEXT_NORMAL_SIZE = MenuItemIdentifier("Make Text Normal Size", "Make Text Normal Size")
        RESTORE_TEXT_AND_SESSION_SIZE = MenuItemIdentifier("Restore Text and Session Size", "Restore Text and Session Size")
        MAKE_TEXT_SMALLER = MenuItemIdentifier("Make Text Smaller", "Make Text Smaller")
        SIZE_CHANGES_UPDATE_PROFILE = MenuItemIdentifier("Size Changes Update Profile", "Size Changes Update Profile")
        START_INSTANT_REPLAY = MenuItemIdentifier("Start Instant Replay", "Start Instant Replay")

    class Session(enum.Enum):
        EDIT_SESSION = MenuItemIdentifier("Edit Session…", "Edit Session…")
        RUN_COPROCESS = MenuItemIdentifier("Run Coprocess…", "Run Coprocess…")
        STOP_COPROCESS = MenuItemIdentifier("Stop Coprocess", "Stop Coprocess")
        RESTART_SESSION = MenuItemIdentifier("Restart Session", "Restart Session")
        OPEN_AUTOCOMPLETE = MenuItemIdentifier("Open Autocomplete…", "Open Autocomplete…")
        OPEN_COMMAND_HISTORY = MenuItemIdentifier("Open Command History…", "Open Command History…")
        OPEN_RECENT_DIRECTORIES = MenuItemIdentifier("Open Recent Directories…", "Open Recent Directories…")
        OPEN_PASTE_HISTORY = MenuItemIdentifier("Open Paste History…", "Open Paste History…")

        class Triggers(enum.Enum):
            ADD_TRIGGER = MenuItemIdentifier("Add Trigger…", "Add Trigger")
            EDIT_TRIGGERS = MenuItemIdentifier("Edit Triggers", "Edit Triggers")
            ENABLE_TRIGGERS_IN_INTERACTIVE_APPS = MenuItemIdentifier("Enable Triggers in Interactive Apps", "Enable Triggers in Interactive Apps")
            ENABLE_ALL = MenuItemIdentifier("Enable All", "Triggers.Enable All")
            DISABLE_ALL = MenuItemIdentifier("Disable All", "Triggers.Disable All")

        RESET = MenuItemIdentifier("Reset", "Reset")
        RESET_CHARACTER_SET = MenuItemIdentifier("Reset Character Set", "Reset Character Set")

        class Log(enum.Enum):
            LOG_TO_FILE = MenuItemIdentifier("Log to File", "Log.Toggle")
            IMPORT_RECORDING = MenuItemIdentifier("Import Recording", "Log.ImportRecording")
            EXPORT_RECORDING = MenuItemIdentifier("Export Recording", "Log.ExportRecording")
            SAVE_CONTENTS = MenuItemIdentifier("Save Contents…", "Log.SaveContents")  #: Deprecated - this has moved elsewhere.


        class TerminalState(enum.Enum):
            ALTERNATE_SCREEN = MenuItemIdentifier("Alternate Screen", "Alternate Screen")
            FOCUS_REPORTING = MenuItemIdentifier("Focus Reporting", "Focus Reporting")
            MOUSE_REPORTING = MenuItemIdentifier("Mouse Reporting", "Mouse Reporting")
            PASTE_BRACKETING = MenuItemIdentifier("Paste Bracketing", "Paste Bracketing")
            APPLICATION_CURSOR = MenuItemIdentifier("Application Cursor", "Application Cursor")
            APPLICATION_KEYPAD = MenuItemIdentifier("Application Keypad", "Application Keypad")
            STANDARD_KEY_REPORTING_MODE = MenuItemIdentifier("Standard Key Reporting Mode", "Terminal State.Standard Key Reporting")
            MODIFYOTHERKEYS_MODE_1 = MenuItemIdentifier("modifyOtherKeys Mode 1", "Terminal State.Report Modifiers like xterm 1")
            MODIFYOTHERKEYS_MODE_2 = MenuItemIdentifier("modifyOtherKeys Mode 2", "Terminal State.Report Modifiers like xterm 2")
            CSI_U_MODE = MenuItemIdentifier("CSI u Mode", "Terminal State.Report Modifiers with CSI u")
            RAW_KEY_REPORTING_MODE = MenuItemIdentifier("Raw Key Reporting Mode", "Terminal State.Raw Key Reporting")
            RESET = MenuItemIdentifier("Reset", "Reset Terminal State")

        BURY_SESSION = MenuItemIdentifier("Bury Session", "Bury Session")

    class Scripts(enum.Enum):
        class Manage(enum.Enum):
            NEW_PYTHON_SCRIPT = MenuItemIdentifier("New Python Script", "New Python Script")
            OPEN_PYTHON_REPL = MenuItemIdentifier("Open Python REPL", "Open Interactive Window")
            MANAGE_DEPENDENCIES = MenuItemIdentifier("Manage Dependencies…", "Manage Dependencies")
            INSTALL_PYTHON_RUNTIME = MenuItemIdentifier("Install Python Runtime", "Install Python Runtime")
            REVEAL_SCRIPTS_IN_FINDER = MenuItemIdentifier("Reveal Scripts in Finder", "Reveal in Finder")
            IMPORT = MenuItemIdentifier("Import…", "Import Script")
            EXPORT = MenuItemIdentifier("Export…", "Export Script")
            CONSOLE = MenuItemIdentifier("Console", "Script Console")


    class Profiles(enum.Enum):
        OPEN_PROFILES = MenuItemIdentifier("Open Profiles…", "Open Profiles…")
        PRESS_OPTION_FOR_NEW_WINDOW = MenuItemIdentifier("Press Option for New Window", "Press Option for New Window")
        OPEN_IN_NEW_WINDOW = MenuItemIdentifier("Open In New Window", "Open In New Window")

    class Toolbelt(enum.Enum):
        SHOW_TOOLBELT = MenuItemIdentifier("Show Toolbelt", "Show Toolbelt")
        SET_DEFAULT_WIDTH = MenuItemIdentifier("Set Default Width", "Set Default Width")

    class Window(enum.Enum):
        MINIMIZE = MenuItemIdentifier("Minimize", "Minimize")
        ZOOM = MenuItemIdentifier("Zoom", "Zoom")
        EDIT_TAB_TITLE = MenuItemIdentifier("Edit Tab Title", "Edit Tab Title")
        EDIT_WINDOW_TITLE = MenuItemIdentifier("Edit Window Title", "Edit Window Title")

        class WindowStyle(enum.Enum):
            NORMAL = MenuItemIdentifier("Normal", "Window Style.Normal")
            FULL_SCREEN = MenuItemIdentifier("Full Screen", "Window Style.Full Screen")
            MAXIMIZED = MenuItemIdentifier("Maximized", "Window Style.Maximized")
            NO_TITLE_BAR = MenuItemIdentifier("No Title Bar", "Window Style.No Title Bar")
            FULLWIDTH_BOTTOM_OF_SCREEN = MenuItemIdentifier("Full-Width Bottom of Screen", "Window Style.FullWidth Bottom of Screen")
            FULLWIDTH_TOP_OF_SCREEN = MenuItemIdentifier("Full-Width Top of Screen", "Window Style.FullWidth Top of Screen")
            FULLHEIGHT_LEFT_OF_SCREEN = MenuItemIdentifier("Full-Height Left of Screen", "Window Style..FullHeight Left of Screen")
            FULLHEIGHT_RIGHT_OF_SCREEN = MenuItemIdentifier("Full-Height Right of Screen", "Window Style.FullHeight Right of Screen")
            BOTTOM_OF_SCREEN = MenuItemIdentifier("Bottom of Screen", "Window Style.Bottom of Screen")
            TOP_OF_SCREEN = MenuItemIdentifier("Top of Screen", "Window Style.Top of Screen")
            LEFT_OF_SCREEN = MenuItemIdentifier("Left of Screen", "Window Style.Left of Screen")
            RIGHT_OF_SCREEN = MenuItemIdentifier("Right of Screen", "Window Style.Right of Screen")

        MERGE_ALL_WINDOWS = MenuItemIdentifier("Merge All Windows", "Merge All Windows")
        ARRANGE_WINDOWS_HORIZONTALLY = MenuItemIdentifier("Arrange Windows Horizontally", "Arrange Windows Horizontally")
        ARRANGE_SPLIT_PANES_EVENLY = MenuItemIdentifier("Arrange Split Panes Evenly", "Arrange Split Panes Evenly")
        MOVE_SESSION_TO_WINDOW = MenuItemIdentifier("Move Session to Window", "Move Session to Window")
        SAVE_WINDOW_ARRANGEMENT = MenuItemIdentifier("Save Window Arrangement", "Save Window Arrangement")
        SAVE_CURRENT_WINDOW_AS_ARRANGEMENT = MenuItemIdentifier("Save Current Window as Arrangement", "Save Current Window as Arrangement")

        class SelectSplitPane(enum.Enum):
            SELECT_PANE_ABOVE = MenuItemIdentifier("Select Pane Above", "Select Split Pane.Select Pane Above")
            SELECT_PANE_BELOW = MenuItemIdentifier("Select Pane Below", "Select Split Pane.Select Pane Below")
            SELECT_PANE_LEFT = MenuItemIdentifier("Select Pane Left", "Select Split Pane.Select Pane Left")
            SELECT_PANE_RIGHT = MenuItemIdentifier("Select Pane Right", "Select Split Pane.Select Pane Right")
            NEXT_PANE = MenuItemIdentifier("Next Pane", "Select Split Pane.Next Pane")
            PREVIOUS_PANE = MenuItemIdentifier("Previous Pane", "Select Split Pane.Previous Pane")


        class ResizeSplitPane(enum.Enum):
            MOVE_DIVIDER_UP = MenuItemIdentifier("Move Divider Up", "Resize Split Pane.Move Divider Up")
            MOVE_DIVIDER_DOWN = MenuItemIdentifier("Move Divider Down", "Resize Split Pane.Move Divider Down")
            MOVE_DIVIDER_LEFT = MenuItemIdentifier("Move Divider Left", "Resize Split Pane.Move Divider Left")
            MOVE_DIVIDER_RIGHT = MenuItemIdentifier("Move Divider Right", "Resize Split Pane.Move Divider Right")


        class ResizeWindow(enum.Enum):
            DECREASE_HEIGHT = MenuItemIdentifier("Decrease Height", "Resize Window.Decrease Height")
            INCREASE_HEIGHT = MenuItemIdentifier("Increase Height", "Resize Window.Increase Height")
            DECREASE_WIDTH = MenuItemIdentifier("Decrease Width", "Resize Window.Decrease Width")
            INCREASE_WIDTH = MenuItemIdentifier("Increase Width", "Resize Window.Increase Width")

        SELECT_NEXT_TAB = MenuItemIdentifier("Select Next Tab", "Select Next Tab")
        SELECT_PREVIOUS_TAB = MenuItemIdentifier("Select Previous Tab", "Select Previous Tab")
        MOVE_TAB_LEFT = MenuItemIdentifier("Move Tab Left", "Move Tab Left")
        MOVE_TAB_RIGHT = MenuItemIdentifier("Move Tab Right", "Move Tab Right")
        PASSWORD_MANAGER = MenuItemIdentifier("Password Manager", "Password Manager")
        PIN_HOTKEY_WINDOW = MenuItemIdentifier("Pin Hotkey Window", "Pin Hotkey Window")
        BRING_ALL_TO_FRONT = MenuItemIdentifier("Bring All To Front", "Bring All To Front")

    class Help(enum.Enum):
        ITERM2_HELP = MenuItemIdentifier("iTerm2 Help", "iTerm2 Help")
        COPY_MODE_SHORTCUTS = MenuItemIdentifier("Copy Mode Shortcuts", "Copy Mode Shortcuts")
        OPEN_SOURCE_LICENSES = MenuItemIdentifier("Open Source Licenses", "Open Source Licenses")
        GPU_RENDERER_AVAILABILITY = MenuItemIdentifier("GPU Renderer Availability", "GPU Renderer Availability")




"""Provides hooks for session life-cycle events."""
import asyncio

import iterm2.connection
import iterm2.notifications


class EachSessionOnceMonitor:
    """
    This is a convenient way to do something to all sessions exactly once,
    including those created in the future.

    You can use it as a context manager to get the session_id of each session,
    or you can use the static method `async_foreach_session_create_task` to
    have a task created for each session.

    :param connection: The :class:`~iterm2.connection.Connection` to use.
    :param app: An instance of :class:`~iterm2.app.App`.
    """
    def __init__(self, app: 'iterm2.app.App'):
        self.__connection = app.connection
        self.__app = app
        self.__token = None
        self.__queue: asyncio.Queue = asyncio.Queue(
            loop=asyncio.get_event_loop())

    @staticmethod
    async def async_foreach_session_create_task(app, task):
        """
        Create a task for each session. Cancels the task when the session
        terminates.

        Includes sessions in existence now and those created in the future.

        :param app: An instance of :class:`~iterm2.app.App`.
        :param task: A coro taking a single argument of session ID.
        :returns: A future.

        Example:

          .. code-block:: python

            app = await iterm2.async_get_app(connection)
            # Print a message to stdout when there's a new prompt in any
            # session
            async def my_task(session_id):
                async with iterm2.PromptMonitor(connection, session_id) as mon:
                    await mon.async_get()
                    print("Prompt detected")

            await (iterm2.EachSessionOnceMonitor.
                async_foreach_session_create_task(app, my_task))
        """
        tasks = {}

        async def each_mon():
            async with EachSessionOnceMonitor(app) as mon:
                while True:
                    session_id = await mon.async_get()
                    tasks[session_id] = asyncio.create_task(task(session_id))

        async def termination_mon():
            async with SessionTerminationMonitor(app.connection) as mon:
                while True:
                    session_id = await mon.async_get()
                    if session_id in tasks:
                        task = tasks[session_id]
                        del tasks[session_id]
                        task.cancel()
        await asyncio.gather(each_mon(), termination_mon())

    async def __aenter__(self):
        async def callback(_connection, message):
            """Called when a new session is created."""
            await self.__queue.put(message)

        self.__token = (
            await iterm2.notifications.
            async_subscribe_to_new_session_notification(
                self.__connection,
                callback))

        for window in self.__app.terminal_windows:
            for tab in window.tabs:
                for session in tab.all_sessions:
                    await self.__queue.put(session)

        return self

    async def async_get(self) -> str:
        """Returns the session ID."""
        result = await self.__queue.get()
        session_id = result.session_id
        return session_id

    async def __aexit__(self, exc_type, exc, _tb):
        try:
            await iterm2.notifications.async_unsubscribe(
                self.__connection, self.__token)
        except iterm2.notifications.SubscriptionException:
            pass


class SessionTerminationMonitor:
    """
    Watches for session termination.

    A session is said to terminate when its command (typically `login`) has
    exited. If the user closes a window, tab, or split pane they can still undo
    closing it for some amount of time. Session termination will be delayed
    until it is no longer undoable.

    :param connection: The :class:`~iterm2.connection.Connection` to use.

    Example:

      .. code-block:: python

          async with iterm2.SessionTerminationMonitor(connection) as mon:
              while True:
                  session_id = await mon.async_get()
                  print("Session {} closed".format(session_id))
    """
    def __init__(self, connection: iterm2.connection.Connection):
        self.__connection = connection
        self.__token = None
        self.__queue: asyncio.Queue = asyncio.Queue(
            loop=asyncio.get_event_loop())

    async def __aenter__(self):
        async def callback(_connection, message):
            """Called when a session terminates."""
            await self.__queue.put(message.session_id)

        self.__token = (
            await iterm2.notifications.
            async_subscribe_to_terminate_session_notification(
                self.__connection,
                callback))
        return self

    async def async_get(self) -> str:
        """
        Returns the `session_id` of a just-terminated session.
        """
        session_id = await self.__queue.get()
        return session_id

    async def __aexit__(self, exc_type, exc, _tb):
        try:
            await iterm2.notifications.async_unsubscribe(
                self.__connection,
                self.__token)
        except iterm2.notifications.SubscriptionException:
            pass


class LayoutChangeMonitor:
    """
    Watches for changes to the composition of sessions, tabs, and windows.

    :param connection: The :class:`~iterm2.connection.Connection` to use.
    """
    def __init__(self, connection: iterm2.Connection):
        self.__connection = connection
        self.__token = None
        self.__queue: asyncio.Queue = asyncio.Queue(
            loop=asyncio.get_event_loop())

    async def __aenter__(self):
        async def callback(_connection, message):
            """Called when the layout changes."""
            await self.__queue.put(message)

        self.__token = (
            await iterm2.notifications.
            async_subscribe_to_layout_change_notification(
                self.__connection, callback))
        return self

    async def async_get(self):
        """
        Blocks until the layout changes.

        Will block until any of the following occurs:

        * A session moves from one tab to another (including moving into its
          own window).
        * The relative position of sessions within a tab changes.
        * A tab moves from one window to another.
        * The order of tabs within a window changes.
        * A session is buried or disintered.

        Use :class:`~iterm2.App` to examine the updated application state.

       Example:

       .. code-block:: python

           async with iterm2.LayoutChangeMonitor(connection) as mon:
               while True:
                   await mon.async_get()
                   print("layout changed")

        """
        await self.__queue.get()

    async def __aexit__(self, exc_type, exc, _tb):
        try:
            await iterm2.notifications.async_unsubscribe(
                self.__connection, self.__token)
        except iterm2.notifications.SubscriptionException:
            pass


class NewSessionMonitor:
    """Watches for the creation of new sessions.

      :param connection: The :class:`~iterm2.connection.Connection` to use.

      .. seealso::
          * Example ":ref:`colorhost_example`"
          * Example ":ref:`random_color_example`"

      Example:

      .. code-block:: python

          async with iterm2.NewSessionMonitor(connection) as mon:
              while True:
                  session_id = await mon.async_get()
                  print("Session ID {} created".format(session_id))

        .. seealso::
            * Example ":ref:`autoalert`"
      """
    def __init__(self, connection: iterm2.Connection):
        self.__connection = connection
        self.__token = None
        self.__queue: asyncio.Queue = asyncio.Queue(
            loop=asyncio.get_event_loop())

    async def __aenter__(self):
        async def callback(_connection, message):
            """Called when a new session is created."""
            await self.__queue.put(message)

        self.__token = (
            await iterm2.notifications.
            async_subscribe_to_new_session_notification(
                self.__connection,
                callback))
        return self

    async def async_get(self) -> str:
        """Returns the new session ID."""
        result = await self.__queue.get()
        session_id = result.session_id
        return session_id

    async def __aexit__(self, exc_type, exc, _tb):
        try:
            await iterm2.notifications.async_unsubscribe(
                self.__connection, self.__token)
        except iterm2.notifications.SubscriptionException:
            pass



"""Provides information about the shell prompt."""
import asyncio
import enum
import typing

import iterm2.api_pb2
import iterm2.capabilities
import iterm2.connection
import iterm2.notifications
import iterm2.rpc

class PromptState(enum.Enum):
    """Describes the states that a command prompt can take."""
    UNKNOWN = -1  #: This version of iTerm2 does not report prompt state (you should upgrade)
    EDITING = 0  #: User is editing the command at the prompt
    RUNNING = 1  #: The last entered command is still executing, and has not finished yet.
    FINISHED = 3  #: The last entered command has finished but there hasn't been a new prompt yet (rare).


class Prompt:
    """Describes a command prompt.

    Shell Integration must be installed for this to work properly.  Do not
    construct this object yourself. Use :func:`~async_get_last_prompt` to get
    an instance.
    """
    def __init__(self, proto):
        self.__proto = proto

    @property
    def prompt_range(self) -> iterm2.util.CoordRange:
        """Gives the :class:`~iterm2.util.CoordRange` of a shell prompt."""
        return iterm2.util.CoordRange.from_proto(self.__proto.prompt_range)

    @property
    def command_range(self) -> iterm2.util.CoordRange:
        """
        Gives the :class:`~iterm2.util.CoordRange` of the command following the
        shell prompt.
        """
        return iterm2.util.CoordRange.from_proto(self.__proto.command_range)

    @property
    def output_range(self) -> iterm2.util.CoordRange:
        """
        Gives the :class:`~iterm2.util.CoordRange` of the output of a command
        following a shell prompt.
        """
        return iterm2.util.CoordRange.from_proto(self.__proto.output_range)

    @property
    def working_directory(self) -> typing.Union[None, str]:
        """Returns the working directory at the time the prompt was printed."""
        return self.__proto.working_directory

    @property
    def command(self) -> typing.Union[None, str]:
        """Returns the command entered at the prompt."""
        return self.__proto.command

    @property
    def state(self) -> PromptState:
        """Returns the state of this command prompt."""
        if not self.__proto.HasField("prompt_state"):
            return PromptState.UNKNOWN
        return PromptState(self.__proto.prompt_state)

    @property
    def unique_id(self) -> typing.Optional[str]:
        """Returns the unique ID of this command prompt.

        Will be None if not available because the version of iTerm2 is too
        old to report it."""
        if self.__proto.HasField("unique_prompt_id"):
            return self.__proto.unique_prompt_id
        return None

async def async_get_last_prompt(
        connection: iterm2.connection.Connection,
        session_id: str) -> typing.Union[None, Prompt]:
    """
    Fetches info about the last prompt in a session.

    :param connection: The connection to iTerm2.
    :param session_id: The session ID for which to fetch the most recent
        prompt.

    :returns: The prompt if one exists, or else `None`.

    :throws: :class:`RPCException` if something goes wrong.
    """
    response = await iterm2.rpc.async_get_prompt(connection, session_id)
    status = response.get_prompt_response.status
    # pylint: disable=no-member
    if status == iterm2.api_pb2.GetPromptResponse.Status.Value("OK"):
        return Prompt(response.get_prompt_response)
    if status == iterm2.api_pb2.GetPromptResponse.Status.Value(
            "PROMPT_UNAVAILABLE"):
        return None
    raise iterm2.rpc.RPCException(
        iterm2.api_pb2.GetPromptResponse.Status.Name(status))

async def async_get_prompt_by_id(
        connection: iterm2.connection.Connection,
        session_id: str,
        prompt_unique_id: str) -> typing.Optional[Prompt]:
    """
    Fetches a Prompt by its unique ID.

    :param connection: The connection to iTerm2.
    :param session_id: The Session ID the prompt belongs to.
    :param prompt_unique_id: The unique ID of the prompt.

    :returns: The prompt if one exists or else `None`.

    :throws: :class:`RPCException` if something goes wrong.
    """
    iterm2.capabilities.check_supports_prompt_id(connection)
    response = await iterm2.rpc.async_get_prompt(
        connection, session_id, prompt_unique_id)
    status = response.get_prompt_response.status
    # pylint: disable=no-member
    if status == iterm2.api_pb2.GetPromptResponse.Status.Value("OK"):
        return Prompt(response.get_prompt_response)
    if status == iterm2.api_pb2.GetPromptResponse.Status.Value(
            "PROMPT_UNAVAILABLE"):
        return None
    raise iterm2.rpc.RPCException(
        iterm2.api_pb2.GetPromptResponse.Status.Name(status))

async def async_list_prompts(
    connection: iterm2.connection.Connection,
    session_id: str,
    first: typing.Optional[str] = None,
    last: typing.Optional[str] = None) -> typing.Optional[Prompt]:
    """
    Fetches a list of prompt unique IDs in a session.

    :param connection: The connection to iTerm2.
    :param session_id: The Session ID the prompt belongs to.
    :param first: If not None, list no prompts before the one with
         this unique ID.
    :param last: If not None, list no prompts after the one with
         this unique ID.
    :returns: List of prompt IDs.

    :throws: :class:`RPCException` if something goes wrong.
    """
    iterm2.capabilities.check_supports_prompt_id(connection)
    response = await iterm2.rpc.async_list_prompts(
        connection, session_id, first, last)
    status = response.list_prompts_response.status
    # pylint: disable=no-member
    if status == iterm2.api_pb2.ListPromptsResponse.Status.Value("OK"):
        return response.list_prompts_response.unique_prompt_id

    raise iterm2.rpc.RPCException(
        iterm2.api_pb2.GetPromptResponse.Status.Name(status))

class PromptMonitor:
    """
    An asyncio context manager to watch for changes to the prompt.

    This requires shell integration or prompt-detecting triggers to be
    installed for prompt detection.

    Note: Older versions of the runtime do not support modes other than PROMPT.
    Attempting to use a mode other than PROMPT when connected to a too-old
    version of iTerm2 will result in an exception telling the user to upgrade.

    :param connection: The :class:`~iterm2.connection.Connection` to use.
    :param session_id: The string session ID to monitor.

    Example:

      .. code-block:: python

          async with iterm2.PromptMonitor(
              connection, my_session.session_id) as mon:
                  while True:
                      await mon.async_get()
                      DoSomething()
    """
    class Mode(enum.Enum):
        """The mode for a prompt monitor."""

        # pylint: disable=line-too-long
        PROMPT = iterm2.api_pb2.PromptMonitorMode.Value("PROMPT")  #: Notify when prompt detected
        COMMAND_START = iterm2.api_pb2.PromptMonitorMode.Value("COMMAND_START")  #: Notify when a command begins execution
        COMMAND_END = iterm2.api_pb2.PromptMonitorMode.Value("COMMAND_END")  #: Notify when a command finishes execution
        # pylint: enable=line-too-long

    def __init__(
            self,
            connection: iterm2.connection.Connection,
            session_id: str,
            modes: typing.Optional[typing.List[Mode]] = None):
        if modes is None:
            modes = [PromptMonitor.Mode.PROMPT]
        self.connection = connection
        self.session_id = session_id
        self.__modes = modes
        self.__token = None
        self.__queue: asyncio.Queue = asyncio.Queue(
            loop=asyncio.get_event_loop())
        if (modes != [PromptMonitor.Mode.PROMPT] and
                not iterm2.capabilities.supports_prompt_monitor_modes(
                    connection)):
            raise iterm2.capabilities.AppVersionTooOld(
                "This version of iTerm2 is too old to handle the " +
                "requested prompt monitor modes (only PROMPT is " +
                "supported in this version). You should upgrade to " +
                "run this script.")

    async def __aenter__(self):
        async def callback(_connection, message):
            """Called when a new prompt is shown."""
            await self.__queue.put(message)

        self.__token = (
            await iterm2.notifications.async_subscribe_to_prompt_notification(
                self.connection,
                callback,
                self.session_id,
                list(map(lambda x: x.value, self.__modes))))
        return self

    async def async_get(
            self,
            include_id: bool = False) -> typing.Union[
                typing.Tuple['PromptMonitor.Mode', typing.Any],
                typing.Tuple['PromptMonitor.Mode', typing.Any, typing.Optional[str]]]:
        """Blocks until a new shell prompt is received.

        Note: Older versions of the runtime that do not support modes other
        than PROMPT always return None.

        :param include_id: If True, return a triple where the last value is the
            prompt's unique ID if available or None otherwise.

        :returns: A tuple of (PROMPT,Optional[Prompt]), (COMMAND_START,Str), or
            (COMMAND_END,Int) where the Str gives the command being run and the
            Int gives the exit status of the command. Older versions of iTerm2 will
            not provide a Prompt object when the first value is PROMPT."""
        triple = await self._async_get()
        if include_id:
            return triple
        return (triple[0], triple[1])

    async def _async_get(self) -> typing.Tuple['PromptMonitor.Mode', typing.Any]:
        message = await self.__queue.get()
        if not iterm2.capabilities.supports_prompt_monitor_modes(
                self.connection):
            return (PromptMonitor.Mode.PROMPT, None, None)
        which = message.WhichOneof('event')
        if which == 'prompt' or which is None:
            if message.prompt.HasField('prompt'):
                prompt = Prompt(message.prompt.prompt)
                return (PromptMonitor.Mode.PROMPT, prompt, message.unique_prompt_id)
            else:
                return (PromptMonitor.Mode.PROMPT, None, message.unique_prompt_id)
        if which == 'command_start':
            return (PromptMonitor.Mode.COMMAND_START,
                    message.command_start.command, message.unique_prompt_id)
        if which == 'command_end':
            return (PromptMonitor.Mode.COMMAND_END, message.command_end.status,
                message.unique_prompt_id)
        raise iterm2.rpc.RPCException(
            f'Unexpected oneof in prompt notification: {message}')

    async def __aexit__(self, exc_type, exc, _tb):
        try:
            await iterm2.notifications.async_unsubscribe(
                self.connection, self.__token)
        except iterm2.notifications.SubscriptionException:
            pass



"""
Helps determine which features are available for the currently connected app.
"""

class AppVersionTooOld(Exception):
    """Raised when you need a newer version of iTerm2."""


# pylint: disable=invalid-name
def ge(a, b):
    """Is a >= b?"""
    if a[0] > b[0]:
        return True
    if a[0] < b[0]:
        return False
    return a[1] >= b[1]
# pylint: enable=invalid-name


def supports_multiple_set_profile_properties(connection):
    """Can multiple profile properties be set in one call?"""
    min_ver = (0, 69)
    return ge(connection.iterm2_protocol_version, min_ver)


def supports_select_pane_in_direction(connection):
    """Can you select pane left/right/up/down?"""
    min_ver = (1, 0)
    return ge(connection.iterm2_protocol_version, min_ver)


def supports_prompt_monitor_modes(connection):
    """Can you monitor the prompt in different modes?"""
    min_ver = (1, 1)
    return ge(connection.iterm2_protocol_version, min_ver)


def supports_status_bar_unread_count(connection):
    """Can the status bar show an unread count?"""
    min_ver = (1, 2)
    return ge(connection.iterm2_protocol_version, min_ver)


def supports_coprocesses(connection):
    """Can you manipulate coprocesses?"""
    min_ver = (1, 3)
    return ge(connection.iterm2_protocol_version, min_ver)


def check_supports_coprocesses(connection):
    """Die if you can't manipulate coprocesses."""
    if not supports_coprocesses(connection):
        raise AppVersionTooOld(
            "This version of iTerm2 is too old to control " +
            "coprocesses from a Python script. You should upgrade to " +
            "run this script.")


def supports_get_default_profile(connection):
    """Can you get the default profile?"""
    min_ver = (1, 4)
    return ge(connection.iterm2_protocol_version, min_ver)


def check_supports_get_default_profile(connection):
    """Die if you can't get the default profile."""
    if not supports_get_default_profile(connection):
        raise AppVersionTooOld(
            "This version of iTerm2 is too old to get the " +
            "default profile from a Python script. You should upgrade " +
            "to run this script.")

def supports_prompt_id(connection):
    """Can you list prompts or get a prompt by ID?"""
    min_ver = (1, 5)
    return ge(connection.iterm2_protocol_version, min_ver)

def check_supports_prompt_id(connection):
    """Die if you can't list prompts."""
    if not supports_prompt_id(connection):
        raise AppVersionTooOld(
            "This version of iTerm2 is too old to fetch a list of " +
            "prompts or get a prompt by ID from a Python script. " +
            "You should upgrade to run this script.")

def supports_list_saved_arrangements(connection):
    """Can you get a list of saved arrangements?"""
    min_ver = (1, 6)
    return ge(connection.iterm2_protocol_version, min_ver)


def check_supports_list_saved_arrangements(connection):
    """Die if you can't get a list of saved arrangements."""
    if not supports_list_saved_arrangements(connection):
        raise AppVersionTooOld(
            "This version of iTerm2 is too old to get the " +
            "default profile from a Python script. You should upgrade " +
            "to run this script.")

def supports_context_menu_providers(connection):
    """Can you register a context menu provider?"""
    min_ver = (1, 7)
    return ge(connection.iterm2_protocol_version, min_ver)

def check_supports_context_menu_provider(connection):
    """"Die if context menu providers are not supported."""
    if not supports_context_menu_providers(connection):
        raise AppVersionTooOld(
            "This version of iTerm2 is too old to register a " +
            "context menu provider. You should upgrade to " +
            "run this script.")

def supports_add_annotation(connection):
    """Can you add an annotation?"""
    min_ver = (1, 8)
    return ge(connection.iterm2_protocol_version, min_ver)

def check_supports_add_annotation(connection):
    """Die if you can't add an annotation."""
    if not supports_add_annotation(connection):
        raise AppVersionTooOld(
            "This version of iTerm2 is too old to add an annotation. " +
            "You should upgrade to run this script.")

def supports_advanced_key_notifications(connection):
  """Can you get key-up and flags-changed notifs?"""
  min_ver = (1, 9)
  return ge(connection.iterm2_protocol_version, min_ver)

def check_supports_advanced_key_notifications(connection):
  """Die if you can't get key-up and flags-changed notifs."""
  if not supports_advanced_key_notifications(connection):
    raise AppVersionTooOld(
        "This version of iTerm2 is too old to get advanced keystroke " +
        "notifications. You should upgrade to run this script.")




import re
from setuptools import setup

def readme():
  with open('README.rst') as f:
    return f.read()

# Load version by parsing _version.py
VERSIONFILE="iterm2/_version.py"
verstrline = open(VERSIONFILE, "rt").read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
if mo:
    verstr = mo.group(1)
else:
    raise RuntimeError("Unable to find version string in %s." % (VERSIONFILE,))


setup(name='iterm2',
      version=verstr,
      description='Python interface to iTerm2\'s scripting API',
      long_description=readme(),
      classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)',
        'Programming Language :: Python :: 3.6',
      ],
      url='http://github.com/gnachman/iTerm2',
      author='George Nachman',
      author_email='gnachman@gmail.com',
      license='GPLv2',
      packages=['iterm2'],
      install_requires=[
          'protobuf',
          'websockets',
          'pyobjc'
      ],
      include_package_data=True,
      zip_safe=False)




"""Manages the details of the websocket connection. """

import asyncio
import os
import sys
import traceback
import typing
import websockets
try:
  import AppKit
  import iterm2.auth
  gAppKitAvailable = True
except:
  gAppKitAvailable = False

# websockets 9.0 moved client into legacy.client and didn't document how to
# migrate to the new API :(. Stick with the old one until I have time to deal
# with this.
try:
  import websockets.legacy.client
  websockets_client = websockets.legacy.client
except:
  websockets_client = websockets.client

import iterm2.api_pb2
from iterm2._version import __version__

def _getenv(key):
    """Gets an environment variable safely.

    Returns None if it does not exist.
    """
    if key in os.environ:
        return os.environ[key]
    return None


def _cookie_and_key():
    cookie = _getenv('ITERM2_COOKIE')
    key = _getenv('ITERM2_KEY')
    return cookie, key


def _headers():
    cookie, key = _cookie_and_key()
    headers = {"origin": "ws://localhost/",
               "x-iterm2-library-version": "python {}".format(__version__),
               "x-iterm2-disable-auth-ui": "true"}
    if cookie is not None:
        headers["x-iterm2-cookie"] = cookie
    elif gAppKitAvailable:
        headers["x-iterm2-advisory-name"] = iterm2.auth.get_script_name()
    if key is not None:
        headers["x-iterm2-key"] = key
    return headers


def _uri():
    return "ws://localhost:1912"


def _subprotocols():
    return ['api.iterm2.com']


class Connection:
    """Represents a loopback network connection from the script to iTerm2.

    Provides functionality for sending and receiving messages. Supports
    dispatching incoming messages."""

    helpers: typing.List[
        typing.Callable[
            ['Connection', typing.Any],
            typing.Coroutine[typing.Any, typing.Any, None]]] = []

    @staticmethod
    def register_helper(helper):
        """
        Registers a function that handles incoming messages.

        You probably don't want to call this. It's used internally for
        dispatching notifications.

        Arguments:
          helper: A coroutine that will be called on incoming messages that
              were not previously handled.
        """
        assert helper is not None
        Connection.helpers.append(helper)

    @staticmethod
    async def async_create() -> 'Connection':
        """Creates a new connection.

        This is intended for use in an apython REPL. It constructs a new
        connection and returns it without creating an asyncio event loop.

        :returns: A new connection to iTerm2.
        """
        connection = Connection()

        # Set ITERM2_COOKIE and ITERM2_KEY if needed by making an Applescript
        # request.
        have_fresh_cookie = connection.authenticate(False)

        while True:
            try:
                connection.websocket = await connection._get_connect_coro()
                # pylint: disable=protected-access
                connection.__dispatch_forever_future = asyncio.ensure_future(
                    connection._async_dispatch_forever(
                        connection, asyncio.get_event_loop()))
                return connection
            except websockets.exceptions.InvalidStatusCode as status_code_exception:
                if status_code_exception.status_code == 401:
                    if have_fresh_cookie:
                        raise
                    # Force request a cookie and try one more time.
                    connection._remove_auth()
                    have_fresh_cookie = connection.authenticate(True)
                    if not have_fresh_cookie:
                        # Didn't get a cookie, so no point trying again.
                        raise
                elif status_code_exception.status_code == 406:
                    print("This version of the iterm2 module is too old for " +
                          "the current version of iTerm2. Please upgrade.")
                    sys.exit(1)
                    raise
                else:
                    raise


    def __init__(self):
        self.websocket = None
        # A list of tuples of (match_func, future). When a message is received
        # each match_func is called with the message as an argument. The first
        # one that returns true gets its future's result set with that message.
        # If none returns True it is dispatched through the helpers. Typically
        # that would be a notification.
        self.__receivers = []
        self.__dispatch_forever_future = None
        self.__tasks = []
        self.loop = None

    def _collect_garbage(self):
        """Asyncio seems to want you to keep a reference to a task that's being
        run with ensure_future. If you don't, it says "task was destroyed but
        it is still pending". So, ok, we'll keep references around until we
        don't need to any more."""
        self.__tasks = list(filter(lambda t: not t.done(), self.__tasks))

    def run_until_complete(self, coro, retry, debug=False):
        """Runs `coro` and returns when it finishes."""
        return self.run(False, coro, retry, debug)

    def run_forever(self, coro, retry, debug=False):
        """Runs `coro` and never returns."""
        self.run(True, coro, retry, debug)

    # pylint: disable=no-self-use
    def set_message_in_future(self, loop, message, future):
        """Sets message as the future's result soon."""
        assert future is not None
        # Is the response to an RPC that is being awaited.

        def set_result():
            """If the future is not done, set its result to message."""
            assert future is not None
            if not future.done():
                future.set_result(message)

        loop.call_soon(set_result)

    async def _async_dispatch_forever(self, connection, loop):
        """
        Read messages from websocket and call helpers or message responders.
        """
        self.__tasks = []
        try:
            while True:
                data = await self.websocket.recv()
                self._collect_garbage()

                message = iterm2.api_pb2.ServerOriginatedMessage()
                message.ParseFromString(data)

                future = self._get_receiver_future(message)
                # Note that however we decide to handle this message,
                # it must be done *after* we await on the websocket.
                # Otherwise we might never get the chance.
                if future is None:
                    # May be a notification.
                    self.__tasks.append(
                        asyncio.ensure_future(
                            self._async_dispatch_to_helper(message)))
                else:
                    self.set_message_in_future(loop, message, future)
        except asyncio.CancelledError:
            # Presumably a run_until_complete script
            pass
        except:
            # I'm not quite sure why this is necessary, but if we don't
            # catch and re-raise the exception it gets swallowed.
            traceback.print_exc()
            raise

    def run(self, forever, coro, retry, debug=False):
        """
        Convenience method to start a program.

        Connects to the API endpoint, begins an asyncio event loop, and runs
        the passed in coroutine. Exceptions will be caught and printed to
        stdout.

        :param forever: Don't terminate after main returns?
        :param coro: A coroutine (async function) to run after connecting.
        :param retry: Keep trying to connect until it succeeds?
        """
        loop = asyncio.get_event_loop()

        async def async_main(connection):
            # Set __tasks here in case coro returns before
            # _async_dispatch_forever starts.
            self.__tasks = []
            dispatch_forever_task = asyncio.ensure_future(
                self._async_dispatch_forever(connection, loop))
            result = await coro(connection)
            if forever:
                await dispatch_forever_task
            dispatch_forever_task.cancel()
            # Make sure the _async_dispatch_to_helper task gets canceled to
            # avoid a warning.
            for task in self.__tasks:
                task.cancel()
            return result

        loop.set_debug(debug)
        self.loop = loop
        return loop.run_until_complete(self.async_connect(async_main, retry))

    async def async_send_message(self, message):
        """
        Sends a message.

        This is a low-level operation that is not generally called by user
        code.

        message: A protocol buffer of type
            iterm2.api_pb2.ClientOriginatedMessage to send.
        """
        await self.websocket.send(message.SerializeToString())

    def _receiver_index(self, message):
        """
        Searches __receivers for the receiver that should handle message and
        returns its index.
        """
        for i in range(len(self.__receivers)):
            match_func = self.__receivers[i][0]
            if match_func and match_func(message):
                return i
        # This says that the first receiver always gets the message if no other
        # receiver can handle it.
        return None

    def _get_receiver_future(self, message):
        """Removes the receiver for message and returns its future."""
        i = self._receiver_index(message)
        if i is None:
            return None
        match_func, future = self.__receivers[i]  # pylint: disable=unused-variable
        del self.__receivers[i]
        return future

    async def async_dispatch_until_id(self, reqid):
        """
        Handle incoming messages until one with the specified id is received.

        Messages not having the expected id get dispatched asynchronously by a
        registered helper if one exists.

        You probably don't want to use this. It's used while waiting for the
        response to an RPC, and has logic specific that that use.

        reqid: The request ID to look for.

        Returns: A message with the specified request id.
        """
        my_future = asyncio.Future()

        def match_func(incoming_message):
            return incoming_message.id == reqid

        my_receiver = (match_func, my_future)
        self.__receivers.append(my_receiver)
        return await my_future

    async def _async_dispatch_to_helper(self, message):
        """
        Dispatch a message to all registered helpers.
        """
        for helper in Connection.helpers:
            # pylint: disable=try-except-raise
            assert helper is not None
            try:
                if await helper(self, message):
                    break
            except Exception:
                raise

    @property
    def iterm2_protocol_version(self):
        """
        Returns a tuple (major version, minor version) or 0,0 if it's an old
        version of iTerm2 that doesn't report its version or it's unknown.
        """
        key = "X-iTerm2-Protocol-Version"
        if key not in self.websocket.response_headers:
            return (0, 0)
        header_value = self.websocket.response_headers[key]
        parts = header_value.split(".")
        if len(parts) != 2:
            return (0, 0)
        return (int(parts[0]), int(parts[1]))

    def _get_connect_coro(self):
        if gAppKitAvailable:
            path = self._unix_domain_socket_path()
            exists = os.path.exists(path)

            if exists:
                return self._get_unix_connect_coro()
        return self._get_tcp_connect_coro()

    def _remove_auth(self):
        # Remove these because they are not re-usable.
        vars = ["ITERM2_COOKIE", "ITERM2_KEY"]
        for var in vars:
            if var in os.environ:
                del os.environ[var]

    def _unix_domain_socket_path(self):
        applicationSupport = os.path.join(
            AppKit.NSSearchPathForDirectoriesInDomains(
                AppKit.NSApplicationSupportDirectory,
                AppKit.NSUserDomainMask,
                True)[0],
            "iTerm2")
        return os.path.join(applicationSupport, "private", "socket")

    def _get_unix_connect_coro(self):
        """Experimental: connect with unix domain socket."""
        path = self._unix_domain_socket_path()
        return websockets_client.unix_connect(
            path,
            "ws://localhost/",
            ping_interval=None,
            extra_headers=_headers(),
            subprotocols=_subprotocols(),
            max_size=None)


    def _get_tcp_connect_coro(self):
        """Legacy: connect with tcp socket."""
        return websockets.connect(_uri(),
                                        ping_interval=None,
                                        extra_headers=_headers(),
                                        subprotocols=_subprotocols())

    def authenticate(self, force):
        """
        Request a cookie via Applescript.

        :param force: Remove existing cookies first?

        :returns: True if a new cookie was gotten. False if not. When `force`
            is True, then a return value of `False` means it wasn't able to
            connect. When `force` is False, a return value of `False` could
            mean either a failure to connect or that there was already a
            (possibly stale) cookie in the environment.
        """
        if not gAppKitAvailable:
            return False
        if force:
            self._remove_auth()
        try:
            return iterm2.auth.authenticate()
        except iterm2.auth.AuthenticationException:
            return False

    async def async_connect(self, coro, retry=False):
        """
        Establishes a websocket connection.

        You probably want to use Connection.run(), which takes care of runloop
        setup for you. Connects to iTerm2 on localhost. Once connected, awaits
        execution of coro.

        This uses ITERM2_COOKIE and ITERM2_KEY environment variables to help
        with authentication. ITERM2_COOKIE has a shared secret that lets
        user-launched scripts skip the auth dialog. ITERM2_KEY is used to tie
        together the output
        of this program with its entry in the scripting console.

        :param coro: A coroutine to run once connected.
        :param retry: Keep trying to connect until it succeeds?
        """
        done = False
        while not done:
            # Set ITERM2_COOKIE and ITERM2_KEY if needed by making an
            # Applescript request. This cookie might be stale, but we'll try it
            # optimstically.
            have_fresh_cookie = self.authenticate(False)

            try:
                async with self._get_connect_coro() as websocket:
                    done = True
                    self.websocket = websocket
                    # pylint: disable=broad-except
                    try:
                        return await coro(self)
                    except Exception as _err:
                        traceback.print_exc()
                        sys.exit(1)
            except websockets.exceptions.InvalidStatusCode as exception:
                if exception.status_code == 401:
                    # Auth failure.
                    if retry:
                        # Sleep and try to authenticate until successful.
                        while not have_fresh_cookie:
                            await asyncio.sleep(0.5)
                            have_fresh_cookie = self.authenticate(True)
                    else:
                        # Not retrying forever.
                        if have_fresh_cookie:
                            # Welp, that shoulda worked. Give up.
                            raise

                        # Prepare the second and final attempt.
                        self._remove_auth()
                        have_fresh_cookie = self.authenticate(True)
                        if not have_fresh_cookie:
                            # Failed to get a cookie. Give up.
                            raise
                elif exception.status_code == 406:
                    print("This version of the iterm2 module is too old " +
                          "for the current version of iTerm2. Please upgrade.")
                    sys.exit(1)
                    raise
                else:
                    raise
            except websockets.exceptions.InvalidMessage:
                # This is a temporary workaround for this issue:
                #
                # https://gitlab.com/gnachman/iterm2/issues/7681#note_163548399
                # https://github.com/aaugustin/websockets/issues/604
                #
                # I'm leaving the print statement in because I'm worried this
                # might have unexpected consequences, as InvalidMessage is
                # certainly not very specific.
                print("websockets.connect failed with InvalidMessage. " +
                      "Retrying.")
            except (ConnectionRefusedError, OSError) as exception:
                # https://github.com/aaugustin/websockets/issues/593
                if retry:
                    await asyncio.sleep(0.5)
                else:
                    print("""
There was a problem connecting to iTerm2.

Please check the following:
  * Ensure the Python API is enabled in iTerm2's preferences
  * Ensure iTerm2 is running
  * Ensure script is running on the same machine as iTerm2

If you'd prefer to retry connecting automatically instead of
raising an exception, pass retry=true to run_until_complete()
or run_forever()

""", file=sys.stderr)
                    if gAppKitAvailable:
                        path = self._unix_domain_socket_path()
                        exists = os.path.exists(path)
                        if exists:
                            print(f"If you have downgraded from iTerm2 3.3.12+ to an older version, you must\nmanually delete the file at {path}.\n", file=sys.stderr)
                    done = True
                    raise
            finally:
                self._remove_auth()


def run_until_complete(
        coro: typing.Callable[[Connection],
                              typing.Coroutine[typing.Any, typing.Any, None]],
        retry=False,
        debug=False) -> None:
    """
    Convenience method to run an async function taking an
    :class:`~iterm2.Connection` as an argument.

    After `coro` returns this function will return.

    :param coro: The coroutine to run. Must be an `async def` function. It
        should take one argument, a :class:`~iterm2.connection.Connection`, and
        does not need to return a value.
    :param retry: Keep trying to connect until it succeeds?
    """
    return Connection().run_until_complete(coro, retry, debug)


def run_forever(
        coro: typing.Callable[[Connection],
                              typing.Coroutine[typing.Any,
                                               typing.Any, None]],
        retry=False,
        debug=False) -> None:
    """
    Convenience method to run an async function taking an
    :class:`~iterm2.Connection` as an argument.

    This function never returns.

    :param coro: The coroutine to run. Must be an `async def` function. It
        should take one argument, a :class:`~iterm2.connection.Connection`, and
        does not need to return a value.
    :param retry: Keep trying to connect until it succeeds?
    """
    Connection().run_forever(coro, retry, debug)
